/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* PARTIPATION FILE reader/writer.
 */

#include "backtalk.h"

#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>

#include "sysdict.h"
#include "user.h"
#include "dict.h"
#include "str.h"
#include "sel.h"
#include "file.h"
#include "ps_part.h"
#include "lock.h"

#ifdef HAVE_LIMITS_H
#include <limits.h>
#else
#define INT_MAX 32767
#endif

#define SECS_PER_DAY (60*60*24)


/* PARTICIPATION FILE - Normally one participation file is kept in memory.
 * It is for the conference in the sysdict variable "conf" and has been
 * read if "part_loaded" is true.
 *
 * "part_modified" is true if any change has been made to the data stored in
 * a PicoSpan-style participation file.  "ext_modified" is true if any of the
 * data added by Backtalk has been changed.  If either has been changed, we
 * should write the new one out before exiting.
 *
 * If "no_part" is true, we were unable to open a participation file for the
 * user (possibly because he is anonymous, possibly because of permission
 * problems) so we will run as if he was just joining for the first time, and
 * not attempt to save any updates.
 *
 * The participation file is stored in a linked list.
 *
 * Note that we hold the participation file open between loading it and
 * writing it, so that our lock on it doesn't go away.
 */

int part_loaded= FALSE;		/* Have we got participation file in memory? */
int part_type;			/* What type of particip. file was it?	 */
				/*   0= Picospan		         */
				/*   1= Picospan plus extension file     */
				/*   2= Backtalk                         */

int part_modified= FALSE;	/* Need to write out part file when done? */
int ext_modified= FALSE;	/* Need to write out extensions when done? */

int snooping= FALSE;		/* Is conference open for snooping only? */

int no_part= FALSE;		/* Don't use a participation file */
time_t part_date= -1L;		/* Last mod date - -1=dunno, 0=brandnew */
FILE *part_fp= NULL;
FILE *ext_fp= NULL;
char part_name[BFSZ];		/* fullpath name of current particip. file */
char *alias= NULL;		/* User's full name, as in particip. file */
HashTable *item_etc;		/* Miscellaneous item fields */

struct pfe {
    int item_no;		/* Item number */
    int resp_no;		/* Number of responses read including item */
    long date;			/* Time of last reading */
    char *my_title;		/* User's own title for the item */
    char forgotten;		/* Is this item forgotten? */
    char favorite;		/* Is this item a favorite? */
    char delete;		/* Should this entry not be written out? */
    HashTable *etc;		/* Other random fields */
    struct pfe *next;		/* Pointer to next entry */
    } *pflist;


void writepart(void);
void freepart(void);
int openpf(int snoop);
int loadpart(int snoop);
void newpart(int);
void writepspart(void);
void writebtpart(int ext);
void loadextpart(void);


/* CLOSEPART -- Close the currently open participation file if there is one,
 * writing it out if it has been modified.  If any item has been modified,
 * it is very important that this be called to write the pf out.
 */

void closepart()
{
    if (!part_loaded) return;

    if ((part_modified || ext_modified) && !snooping)
	writepart();
    else
	freepart();

    if (part_fp != NULL)
    {
	fclose(part_fp);
	part_fp= NULL;
	unlock(part_name);
    }

    pflist= NULL;

    part_loaded= FALSE;
    part_date= -1L;
}


/* REMOVEPART -- Close and delete the participation file of the currently open
 * conference.
 */

void removepart()
{
    if (no_part || !part_loaded) return;

    if (part_fp != NULL)
    {
	fclose(part_fp);
	part_fp= NULL;
	unlock(part_name);
    }

    if (!snooping)
    {
#if defined(PARTUTIL) && !defined(PART_DIR)
	partutil("-d",sdict(VAR_CONF),sdict(VAR_ID));
#else
	unlink(part_name);
#endif
    }

    freepart();
    pflist= NULL;
    part_loaded= part_modified= ext_modified= FALSE;
    no_part= TRUE;
    part_date= -1L;
}


/* OPENPART - Make sure the participation file for the current conference has
 * been loaded in memory.  If the argument is true, we will open it in "snoop"
 * mode, in which participation files are read-only.  You don't normally have
 * to call this unless you want to snoop, because calling most any routine
 * that does anything to a participation file does call this.  Returns true if
 * a new participation file was created.
 */

int openpart(int snoop)
{
    if (part_loaded) return 0;
    snooping= snoop;

    if (sdict(VAR_ID)[0] == '\0' || sdict(VAR_HOMEDIR)[0] == '\0')
    {
	no_part= TRUE;
	return 0;
    }
    else
	return loadpart(snoop);
}


/* LOADPART -- Read in a user's participation file for the currnt conference
 * given directory name.  Create an empty one (in memory, so far) if he
 * doesn't have one.  If snoop is true, we don't create a participation
 * file.  Returns true if we created one.
 */

int loadpart(int snoop)
{
    char buf[BFSZ];
    int len;

#ifdef PART_DIR
    bbspartpath(part_name,sdict(VAR_ID),sdict(VAR_PARTICIP));
#else
    homepartpath(part_name,sdict(VAR_HOMEDIR),sdict(VAR_PARTICIP));
#endif

    /* Open, create, and lock the participation file */
    if (openpf(snoop))
    {
	newpart(!no_part);
	return 1;
    }
    else if (no_part || snoop)
	return 0;

    /* Read and check magic code number */
    if (fgets(buf,BFSZ,part_fp) == NULL)
    {
    	/* Ignore empty file */
	newpart(FALSE);
	return 1;
    }

    if (!strcmp(buf,"!<pr03>\n"))
    {
#if DEFAULT_PART_TYPE == 0
        return loadpspart();
#else
	if (loadpspart()) return 1;
	/* Try loading an extension file */
	loadextpart();
	return 0;
#endif
    }

    if (!strcmp(buf,"!<pr10>\n"))
        return loadbtpart();

    if ((len= strlen(buf)) < 8 &&
        (!strncmp(buf,"!<pr03>",len) || !strncmp(buf,"!<pr10>",len)) )
    {
	/* Ignore truncated participation file */
	newpart(FALSE);
	return 1;
    }

    /* Bomb on things that can't have been part files */
    die("Participation file %s has bad magic number", part_name);
}


/* LOADPSPART -- Read in a Picospan-style participation file.  The file's name
 * is already in the global part_name, and it has already been opened on the
 * part_fp file pointer, and the first line has already been read in and
 * processed.  We just finish the job, returning 0 if we were able to read
 * the file, and 0 if we had to concoct a new one.
 */

int loadpspart()
{
    char buf[BFSZ], *t;
    struct pfe *npf,*lpf=NULL;

    /* Read alias name */
    if (fgets(buf,BFSZ,part_fp) == NULL)
    {
	/* EOF - Assume participation file was truncated */
	newpart(FALSE);
	return 1;
    }
    buf[BFSZ-2]= '\n'; buf[BFSZ-1]= '\0';
    if ((t= strchr(buf,'\n')) == NULL)
    {
	/* No terminating newline - Assume file was truncated */
	newpart(FALSE);
	return 1;
    }
    *t= '\0';
    alias= strdup(buf);
    sets(VAR_ALIAS,buf,SM_COPY);

    while (fgets(buf,BFSZ,part_fp) != NULL)
    {
	if (index(buf,'\n') == NULL)
	    /* Discard lines with no terminating newline */
	    continue;

	npf= (struct pfe *)malloc(sizeof(struct pfe));
#ifdef YAPP_COMPAT
	if (sscanf(buf,"%d %d %lx",&npf->item_no,
				   &npf->resp_no,
				   (unsigned long *)&npf->date) != 3)
	    die("syntax error in participation file %s: %s",part_name,buf);
	if (npf->forgotten= (npf->resp_no < 0))
	    npf->resp_no= -npf->resp_no;
#else
	/* Fancier scanning to handle case of "-0" responses */
	if (!isascii(buf[0]) || !isdigit(buf[0]))
	    die("syntax error in participation file %s: '%s'",part_name,buf);
	npf->item_no= atoi(buf);
	t= firstout(firstin(buf," ")," ");
	if (npf->forgotten= (*t == '-')) t++;
	if (sscanf(t,"%d %lx",&npf->resp_no,
			      (unsigned long *)&npf->date) != 2)
	    die("syntax error in participation file %s: %s",part_name,buf);
#endif /* YAPP_COMPAT */

	npf->delete= 0;
	npf->favorite= 0;
	npf->my_title= NULL;
	npf->etc= NULL;
	npf->next= NULL;
	if (lpf == NULL)
	    pflist= npf;
	else
	    lpf->next= npf;
	lpf= npf;
    }

    part_loaded= TRUE;
    part_modified= FALSE;
    ext_modified= FALSE;
    item_etc= NULL;
    return 0;
}


/* ADD_ETC - Save a value to the item or response "etc" dictionary.
 * Just a helper function for loadbtpart().
 */

void add_etc(HashTable **tp, char *var, int is_string, char *sval, int ival)
{
    HashEntry *h;
    int isnew;

    if (*tp == NULL)
    {
	*tp= (HashTable *)malloc(sizeof(HashTable));
	InitHashTable(*tp);
    }
    h= CreateHashEntry(*tp,var,&isnew);
    if (!isnew) free_val(&(h->t));
    if (is_string)
    {
	h->t.val= (void *)strdup(sval);
	h->t.flag= TK_STRING | TKF_FREE;
    }
    else
    {
	h->t.val= (void *)ival;
	h->t.flag= TK_INTEGER;
    }
}

/* LOADBTPART -- Read in a Backtalk-style participation file.  The file's name
 * is already in the global part_name, and it has already been opened on the
 * part_fp file pointer, and the first line has already been read in and
 * processed.  We just finish the job, returning 0 if we were able to read
 * the file, and 0 if we had to concoct a new one.
 */

int loadbtpart()
{
    char buf[BFSZ], *var, *val, *ch, *save;
    int save_size, save_i;
    long ival;
    struct pfe *npf= NULL, *lpf= NULL;
    int is_string;

    /* Set up our growable buffer for reading in strings */
    save= (char *)malloc(BFSZ);
    save_size= BFSZ;

    while (fgets(buf,BFSZ,part_fp) != NULL)
    {
	/* Lines starting with numbers point to a new item number. */
    	if (isascii(buf[0]) && isdigit(buf[0]))
	{
	    npf= (struct pfe *)malloc(sizeof(struct pfe));

	    npf->item_no= atoi(buf);
	    npf->delete= 0;
	    npf->favorite= 0;
	    npf->forgotten= 0;
	    npf->my_title= NULL;
	    npf->resp_no= 0;
	    npf->date= 0;
	    npf->etc= NULL;
	    npf->next= NULL;
	    continue;
	}

	/* Ignore blanks */
	if (buf[0] == '\n') continue;

	/* Find start of value */
	if ((val= strchr(buf,'=')) == NULL)
	    die("syntax error in participation file %s: %s",part_name,buf);
	*val= '\0';
	*val++;

	/* Save variable name */
	var= strdup(buf);

	is_string= (val[0] == '(');

	if (is_string)
	{
	    int esc= 0;
	    save_i= 0;
	    ch= val+1;
	    for (;;)
	    {
	    	if (ch == '\0')
		{
		    /* Long line broken by fgets or truncated file */

		    if (fgets(buf,BFSZ,part_fp) == NULL)
			break;  /* truncated - stop */
		    
		    ch= buf;
		}

		if (esc= (*ch == '\\' && !esc))
		    continue;

		if (*ch == ')' && !esc);
		    break;	/* Normal end of string */
		
		if (save_i+1 >= save_size)
		{
		    save_size+= BFSZ;
		    save= (char *)realloc(save, save_size);
		}
		save[save_i++]= *ch;
	    }
	    save[save_i]= '\0';
	}
	else if (val[0] == '\n')
	    ival= 1;
	else if (isascii(val[0]) && (isdigit(val[0]) || val[0] == '-'))
	    ival= atoi(val);
	else
	    die("syntax error in participation file %s: %s",part_name,buf);

	/* Handle special fields */
	if (npf == NULL)
	{
	    if (!strcmp(var,"alias"))
	    {
		if (is_string) alias= strdup(save);
	    	free(var);
		continue;
	    }
	    else
		add_etc(&item_etc,var,is_string,save,ival);
	}
	else
	{
	    if (!strcmp(var,"r"))
	    {
	    	/* This gets much fancier in the future */
		if (!is_string) npf->resp_no= ival;
		free(var);
		continue;
	    }
	    else if (!strcmp(var,"d"))
	    {
		if (!is_string) npf->date= ival;
		free(var);
		continue;
	    }
	    else if (!strcmp(var,"forgot"))
	    {
		if (!is_string) npf->forgotten= ival;
		free(var);
		continue;
	    }
	    else if (!strcmp(var,"fav"))
	    {
		if (!is_string) npf->favorite= ival;
		free(var);
		continue;
	    }
	    else if (!strcmp(var,"title"))
	    {
		if (is_string) npf->my_title= strdup(save);
		free(var);
		continue;
	    }
	    else
		add_etc(&(npf->etc),var,is_string,save,ival);
	}

    }
    part_loaded= TRUE;
    part_modified= FALSE;
    ext_modified= FALSE;
    item_etc= NULL;
    return 0;
}


/* NEWPART -- Create a new participation file for the conference in memory.
 * If the "newuser" flag is true, we construct a participation file with some
 * unseen items and some new items, as determined by the "new_isel" and
 * "new_since" settings in the conference settings file.  Otherwise, we
 * just make all items new.
 */

void newpart(int newuser)
{
    pflist= NULL;

    /* Set alias name */
    alias= strdup(user_fname());
    sets(VAR_ALIAS,alias,SM_COPY);

    part_loaded= TRUE;
    part_modified= !no_part;
    ext_modified= FALSE;
    part_type= DEFAULT_PART_TYPE;
    item_etc= NULL;

    if (newuser)
    {
    	char new_isel[BFSZ+1];
    	int new_since, item, nresp, flags;
    	time_t idate, rdate, sdate, now;
	struct pfe *pfn, *pfp;

	part_date= 0L;

	/* Get instructions for which should be new from the settings file  */
	get_settings(sdict(VAR_CONFDIR),new_isel,&new_since);

	/* Everything is new - do nothing */
	if (!strcmp(new_isel,"1-$")) return;

	now= time(0L);
	sdate= now - new_since * SECS_PER_DAY;

	pfn= pfp= NULL;
	for (seek_item(1); !next_item(&item,&nresp,&flags,&idate,&rdate); )
	{
	    pfn= (struct pfe *)malloc(sizeof(struct pfe));
	    pfn->item_no= item;
	    pfn->delete= 0;
	    pfn->forgotten= 0;
	    pfn->favorite= 0;
	    pfn->resp_no= 0;
	    pfn->my_title= NULL;
	    pfn->etc= NULL;

	    if ((new_since > 0 && rdate >= sdate) || in_sel(item,new_isel))
		pfn->date= 0;	/* Item New */
	    else
		pfn->date= now;	/* Item Unseen */

	    pfn->next= NULL;
	    if (pfp == NULL)
	    	pflist= pfn;
	    else
		pfp->next= pfn;
	    pfp= pfn;
	}

	/* Special handling for last item and $ sign in isels */
	if (pfn != NULL && pfn->date != 0 && strchr(new_isel,'$') != NULL)
	    pfn->date= 0;
    }
}


/* LAST_READ -- Return the number of the last response to item that the user
 * has read.  Return resp_no = -1 if it is unread.  The flag forgotten is
 * set to true if the item has been forgotten.
 */

void last_read(int item, int *resp_no, int *forgotten, time_t *date)
{
    struct pfe *pf;
	
    /* Set default values */
    *forgotten= FALSE;
    *resp_no= -1;
    *date= 0;

    openpart(FALSE);

    if (no_part) return;

    for (pf= pflist; pf != NULL; pf= pf->next)
    {
	/* Item not in participation file...is new */
	if (pf->item_no > item)
	    return;
	
	/* If we found it, return the response number */
	if (pf->item_no == item)
	{
	    *forgotten= pf->forgotten;
	    *resp_no= pf->resp_no-1;
	    *date= pf->date;
	    return;
	}
    }
    return;
}


/* UPDATE_ITEM -- Mark an item read up through the given response.  If the
 * response is -1, the entire item is unseen.  If the response is -2, forget
 * the item.  If response is -3, don't update the response number -- only
 * update the time.  If the response is -4, remember (unforget) the item.
 * If the item is forgotten, marking it seen does not unforget it.
 * This version is fancied up so that multiple calls on the same item in
 * rapid succession can be handled efficiently.  Unless rewind is set, the
 * this will only increase the number of responses seen.  If resp is -1, it
 * will not be changed.
 */

struct pfe *last_part= NULL;

void update_item(int item, int resp, time_t now, int rewind)
{
    struct pfe *pf, *pfp, *pfn;
	
    openpart(FALSE);

    if (no_part) return;

    if (last_part != NULL && last_part->item_no == item)
	pf= last_part;
    else
    {
	for (pf= pflist, pfp= NULL; pf != NULL; pfp= pf, pf= pf->next)
	{
	    if (pf->item_no >= item)
		break;
	}

	if (pf == NULL || pf->item_no != item)
	{
	    /* Item not in participation file...put it in */
	    pfn= (struct pfe *)malloc(sizeof(struct pfe));
	    pfn->item_no= item;
	    if (resp == -2)
	    {
	        /* forget an unseen item */
#ifdef YAPP_COMPAT
		pfn->resp_no= 1;
#else
		pfn->resp_no= 0;
#endif
		pfn->forgotten= 1;
		pfn->date= 0;
	    }
	    else
	    {
		pfn->resp_no= (resp < 0) ? 0 : resp+1;
		pfn->forgotten= 0;
		pfn->date= now;
	    }
	    pfn->favorite= 0;
	    pfn->my_title= NULL;
	    pfn->etc= NULL;
	    pfn->delete= 0;
	    pfn->next= pf;
	    if (pfp == NULL)
		pflist= pfn;
	    else
		pfp->next= pfn;
	    part_modified= TRUE;
	    return;
	}
    }

    /* Update existing entry */
    if (resp == -2 || resp == -4)
    {
	/* Remembering and Forgetting */
	if ((resp == -4) == (pf->forgotten)) 
	{
#ifdef YAPP_COMPAT
	    /* Yapp participation file format can't handle forgetting unread
	     * items so mark the first response read in that case.
	     */
	    if (pf->resp_no == 0) pf->resp_no= 1;
#endif
	    pf->forgotten= !pf->forgotten;
	    part_modified= TRUE;
	}
    }
    else
    {
	if (resp != -3 && (rewind || pf->resp_no <= resp))
	    pf->resp_no= resp+1;
	pf->date= now;
	part_modified= TRUE;
    }
    last_part= pf;
}


/* LOST_ITEMS -- The items between first and last are missing.  Delete them from
 * the partipation file.  Really just mark them for deletion.  If last is -1,
 * then delete through the end of the conference.
 */

void lost_items(int first,int last)
{
    struct pfe *pf;
    int state=1;

    if (no_part || last < first) return;

    openpart(FALSE);

    if (no_part) return;

    if (last == -1) last= INT_MAX;
	
    for (pf= pflist; pf != NULL; pf= pf->next)
    {
	if (state == 1 && pf->item_no >= first)
	    state= 2;
		
	if (state == 2)
	{
	    if (pf->item_no > last) break;

#if 0
	    /* If an item was deleted by Picospan, its index file
	     * may still be there.  But the following test means
	     * every user whoever saw the item before it was
	     * deleted would try to delete the index file once.
	     * That's a lot of overhead.  A cron script cleaning
	     * up leftover index files is a better solution.
	     */
	    if (!pf->delete)
		remove_index(sdict(VAR_CONFDIR),pf->item_no);
#endif
	    pf->delete= 1;
	}
    }
}


/* WRITEPART -- Write out the participation file.  A loadpart() or newpart()
 * call must preceed this.  This deallocates the memory that it was stored
 * in as a side effect.
 */

void writepart()
{
    struct pfe *pf, *npf;

    if (no_part)
    {
	if (part_loaded) freepart();
	return;
    }

    if (part_type == 0)
	writepspart();
    else if (part_type == 1)
    {
	if (ext_modified)
	    writebtpart(1);
	if (part_modified)
	    writepspart();
	else
	    freepart();
    }
    else
	writebtpart(0);
}


/* WRITEPSPART -- Write out a PicoSpan-style participation file.  This
 * deallocates the memory that it was stored in as a side effect.
 */

void writepspart()
{
    struct pfe *pf, *npf;
    rewind(part_fp);

    fputs("!<pr03>\n",part_fp);
    fputs(alias,part_fp);
    fputc('\n',part_fp);
    free(alias);

    if (item_etc != NULL)
    {
	DeleteHashTable(item_etc);
	free(item_etc);
	item_etc= NULL;
    }

    for (pf= pflist; pf != NULL; pf= npf)
    {
	if (!pf->delete)
	{
#ifdef YAPP_COMPAT
	    fprintf(part_fp,"%d %d %lX\n",
		    pf->item_no,
		    pf->resp_no * (pf->forgotten ? 1 : -1),
		    (unsigned long)pf->date);
#else
	    /* This version allows a "-0" to be printed, which Picospan
	     * understands just fine, but which Yapp considers freakish.
	     */
	    fprintf(part_fp, pf->forgotten ? "%d -%d %lX\n" : "%d %d %lX\n",
		    pf->item_no,
		    pf->resp_no,
		    (unsigned long)pf->date);
#endif /* YAPP_COMPAT */
	    if (pf->my_title != NULL)
		free(pf->my_title);
	    if (pf->etc != NULL)
	    {
		DeleteHashTable(pf->etc);
		free(pf->etc);
	    }
	}
	npf= pf->next;
	free(pf);
    }

    /* We need this here because the file can get smaller, and we can't
     * destroy and recreate it because we'd lose our lock.
     */

    ftruncate(fileno(part_fp),(size_t)ftell(part_fp));
}


/* WRITEBTPART -- Write out a Backtalk-style participation file.  This
 * deallocates the memory that it was stored in as a side effect.
 *
 * If ext is true, then write an extenstion file instead, and do not deallocate
 * anything.
 */

void writebtpart(int ext)
{
    HashEntry *h;
    HashSearch hs;
    struct pfe *pf,*npf;
    FILE *fp= ext ? part_fp : ext_fp;

    rewind(fp);

    if (ext)
	fputs("!<px01>\n",fp);
    else
    {
	fputs("!<pr10>\n",fp);
	wrtstrtag(fp,"alias",alias);
	free(alias);
    }

    if (item_etc != NULL)
    {
	for (h= FirstHashEntry(item_etc, &hs); h != NULL; h= NextHashEntry(&hs))
	    wrttag(fp, h->key, &(h->t));
	if (!ext)
	{
	    DeleteHashTable(item_etc);
	    free(item_etc);
	    item_etc= NULL;
	}
    }

    for (pf= pflist; pf != NULL; pf= npf)
    {
	if (!pf->delete)
	{
	    if (ext)
	    {
		if (pf->favorite || pf->my_title != NULL || pf->etc != NULL)
		    fprintf(fp,"%d\n",pf->item_no);
	    }
	    else
	    {
		fprintf(fp,"%d\nr=%d\nd=%ld\n",
			pf->item_no,
			pf->resp_no,
			(unsigned long)pf->date);
		if (pf->forgotten) puts("forgot\n");
	    }
	    if (pf->favorite) puts("fav\n");
	    if (pf->my_title != NULL)
	    {
		wrtstrtag(fp, "title", pf->my_title);
		if (!ext) free(pf->my_title);
	    }
	    if (pf->etc != NULL)
	    {
		for (h= FirstHashEntry(pf->etc, &hs); h != NULL;
		     h= NextHashEntry(&hs))
		    wrttag(fp, h->key, &(h->t));
		if (!ext)
		{
		    DeleteHashTable(pf->etc);
		    free(pf->etc);
		}
	    }
	}
	npf= pf->next;
	if (!ext) free(pf);
    }

    /* We need this here because the file can get smaller, and we can't
     * destroy and recreate it because we'd lose our lock.
     */

    ftruncate(fileno(fp),(size_t)ftell(fp));
}


/* FREEPART -- Deallocate the memory we have the participation file stored
 * in.
 */

void freepart()
{
    struct pfe *pf, *npf;

    free(alias);
    if (item_etc != NULL)
    {
	DeleteHashTable(item_etc);
	free(item_etc);
	item_etc= NULL;
    }

    for (pf= pflist; pf != NULL; pf= npf)
    {
	if (pf->my_title != NULL)
	    free(pf->my_title);
	if (pf->etc != NULL)
	{
	    DeleteHashTable(pf->etc);
	    free(pf->etc);
	}
	npf= pf->next;
	free(pf);
    }
}


#ifdef PART_DIR

/* BBSPARTPATH -- Construct the full path name of the participation file,
 * given the file name.  It is saved in the global "part_name" variable.
 * Returns a pointer to "part" -- Yapp 3.0 version.
 */

char *bbspartpath(char *path, char *id, char *name)
{
#if PART_DIR_LEVEL == 0
    sprintf(path,PART_DIR"%s/%s", id, name);
#elif PART_DIR_LEVEL == 1
    sprintf(path,PART_DIR"%c/%s/%s", id[0], id, name);
#else
    sprintf(path,PART_DIR"%c/%c/%s/%s",
      id[0], id[1]=='\0' ? '_' : id[1], id, name);
#endif
    return(path);
}

#endif /* PART_DIR */

/* HOMEPARTPATH -- Construct the full path name of the participation file,
 * given the file name.  It is saved in the global "part_name" variable.
 * Returns a pointer to "part" -- Normal Version, but also used with Yapp 3.0
 * to find .cflist files.
 */

char *homepartpath(char *path, char *dir, char *name)
{
    int homelen;

    strcpy(path,dir);
    homelen= strlen(path);
#ifdef ALLOW_CFDIR
    strcpy(path+homelen,"/.cfdir/");
    if (access(path,0))
	strcpy(path+homelen+1,name);
    else
	strcat(path+homelen,name);
#else
    path[homelen]= '/';
    strcpy(path+homelen+1,name);
#endif /* ALLOW_CFDIR */
    return(path);
}


/* OPENPF - Open the current participation file and store the filepointer in
 * the part_fp global variable.  If there was no previous participation file
 * return 1, otherwise return 0.  If snoop is true, never create the file.
 */

int openpf(int snoop)
{
    no_part= FALSE;

    /* Try opening an existing particiption file */
    part_fp= fopen(part_name, snoop ? "r" : "r+");

#ifdef PART_DIR
    if (!snoop && part_fp == NULL && errno == ENOENT)
    {
	/* User directory missing - create it */
	char *dir1= strdup(part_name);
	char *slash= strrchr(dir1,'/');
	int rc;
	*slash= '\0';
	rc= mkdir(dir1,0700);
#if PART_DIR_LEVEL > 0
	if (rc && errno == ENOENT)
	{
	    /* First hierarchical directory missing - create it */
	    char *dir2= strdup(dir1);
	    slash= strrchr(dir2,'/');
	    *slash= '\0';
	    rc= mkdir(dir2,0700);
#if PART_DIR_LEVEL > 1
	    if (rc && errno == ENOENT)
	    {
		/* Second hierarchical directory missing - create it */
		char *dir3= strdup(dir2);
		slash= strrchr(dir3,'/');
		*slash= '\0';
		mkdir(dir3,0700);
		mkdir(dir2,0700);
		free(dir3);
	    }
#endif /* PART_DIR_LEVEL > 1 */
	    mkdir(dir1,0700);
	    free(dir2);
	}
#endif /* PART_DIR_LEVEL > 0 */
	free(dir1);
	part_fp= fopen(part_name, "r+");
    }
#else

#ifdef PARTUTIL
    /* If we fail because of permissions, try using partutil to fix them */
    if (part_fp == NULL && errno == EACCES)
    {
	partutil("-p",sdict(VAR_CONF),sdict(VAR_ID));
	part_fp= fopen(part_name, snoop ? "r" : "r+");
    }
#endif /* PARTUTIL */
#endif /* !PART_DIR */

    /* If we opened file, lock it and return */
    if (part_fp != NULL)
    {
	if (!no_part)
	    lock_exclusive(fileno(part_fp),part_name);
	return 0;
    }
    else if (snoop)
	return 1;

    /* If file exist but isn't readable then don't try to create it and
     * just run without one.
     */
    if (errno == EACCES)
    {
	no_part= TRUE;
	return 1;
    }

    /* File doesn't exist - create one */
#ifdef PARTUTIL
    partutil("-c",sdict(VAR_CONF),sdict(VAR_ID));
#endif
    no_part= (no_part || ((part_fp= fopen(part_name,"w+")) == NULL));
    if (!no_part)
        lock_exclusive(fileno(part_fp),part_name);
    return 1;
}


#ifdef PARTUTIL

#include "waittype.h"

/* PARTUTIL - execute the partutil program to create, permit, or destory
 * user's participation file for a conference.  Flag should be "-c" to
 * create, "-p" to permit, or "-d" to destroy a participation file.  A flag
 * of "-l" permits (or creates, if necessary) a .cflist for the user.  In
 * this case conf should be NULL.
 */

int partutil(char *flag, char *conf, char *user)
{
    pid_t cpid,wpid;
    waittype status;

    if ((cpid= fork()) == 0)
    {
	/* Child process: do the exec */
	if (conf == NULL)
	    execl(PARTUTIL,"partutil",flag,user,(char *)NULL);
	else
	    execl(PARTUTIL,"partutil",flag,conf,user,(char *)NULL);
	exit(127);
    }

    while ((wpid= wait(&status)) != -1 && wpid != cpid)
	;

    return (WIFEXITED(status)) ? WEXITSTATUS(status) : 127;
}
#endif /* PARTUTIL */


/* SET_CONF_ALIAS - Change the default full name for the user for the
 * currently open conference.  "newname" should be malloc'ed memory.
 * Aliases with newlines are clipped at the newline.
 */

void set_conf_alias(char *newname)
{
    char *p;
    if ((p= strchr(newname,'\n')) != NULL) *p= '\0';
    if (alias != NULL) free(alias);
    alias= newname;
    part_modified= TRUE;
}

/* PARTDATE - Return the last modification date of the currently conference's
 * participation file.  If it didn't exist before this session, return 0.
 */

time_t partdate()
{
    struct stat st;
    if (!part_loaded) return 0L;
    if (part_date == -1L)
	part_date= stat(part_name,&st) ? 0L : st.st_mtime;
    return part_date;
}

/* DUMMIES */
void loadextpart() {}
